import Tokenizer from "./tokenizer.ts";
import ASTGen, { EvalNode, LambdaNode, CallNode, NumberNode, ExpressionNode } from "./ast.ts";
// import Compiler from "./compiler.ts";
const input = Deno.readTextFileSync('code.l')

const tokenizer = new Tokenizer(input);
const tokens = tokenizer.tokenize();

console.log(tokens)

const astGenerator = new ASTGen(tokens);

let ast;
try {
    ast = astGenerator.parse()
} catch (error) {
    console.error(error);
    console.log('at', astGenerator.position, tokens.map((a, i) => i == astGenerator.position ? `${a.type}(${a.value}) <--` : `${a.type}(${a.value})`).join('\n'))
    Deno.exit(1)
}

console.log(ast)

class Lambda {
    argname: string;
    definition: ExpressionNode | Lambda | NumberNode;
    constructor(argn: string, def: ExpressionNode | LambdaNode | NumberNode) {
        this.argname = argn;
        if (def.type == 'Lambda') {
            this.definition = new Lambda(def.arg.value, def.definition)
        } else this.definition = def;
    }
    replace(arg: string, value: string): void {
        console.log(`replaceing`, arg, 'with', value)
        if (this.definition instanceof Lambda) {
            return this.definition.replace(arg, value)
        } else if (this.definition.type == 'Expression') {
            const expr = this.definition as ExpressionNode;
            expr.expression = expr.expression.replaceAll(arg, value)
        }
    }
    eval(value: string): ExpressionNode | NumberNode {
        // console.log('eval thing', value)
        if (this.definition instanceof Lambda) {
            return this.definition.eval(value)
        } else {
            this.replace(this.argname, value);
            return this.definition
        }

    }
}

class Call {
    lambda: Lambda | Call;
    arg: ExpressionNode | NumberNode;
    constructor(lambda: LambdaNode | CallNode, arg: ExpressionNode | NumberNode | CallNode) {
        if (lambda.type == 'Lambda') {
            this.lambda = new Lambda(lambda.arg.value, lambda.definition)
        } else {
            this.lambda = new Call(lambda.function, lambda.arg)
        }
        if (arg.type == 'Call') {
            this.arg = new Call(arg.function, arg.arg).eval()
        } /*else if (arg.type == 'Lambda') 
            this.arg = new Lambda(arg.arg.value, arg.definition)*/
            else this.arg = arg;
    }
    reduce(): ExpressionNode | NumberNode {
        if (this.lambda instanceof Lambda){
            const reduced = this.lambda.eval(
                this.arg.type == 'Expression' ?
                    (this.arg as ExpressionNode).expression
                    : String((this.arg as NumberNode).value)
                )
            return reduced
        } else return this.lambda.reduce()
    }
    ev(): Call {
        this.eval();
        return this
    }
    eval(): ExpressionNode | NumberNode {
        console.log('evaluating', this)
        if (this.lambda instanceof Call) {
            console.debug('sufsiunaifsun', this.lambda?.lambda?.definition?.definition.expression)
            this.lambda.eval()
            console.debug('1232132132121', this.lambda?.lambda?.definition?.definition.expression, this.arg, 'soskosafm', this.lambda)
            // this.lambda.eval()
            // const res = this.ev();
            console.log('result a:', res, this.arg)
            throw res.lambda;
        }
        console.log('result b:', this.reduce())
        return this.reduce()
    }
}

for (const node of ast) {
    if (node.type != 'Eval') continue;
    const evalNode = node as EvalNode;
    const call = new Call(evalNode.call.function, evalNode.call.arg);
    console.log(call.eval())
}

// const compiler = new Compiler(ast);

// for (const node of compiler.AST) {
//     compiler.compile(node)
// }

Deno.writeTextFileSync('ast.json', JSON.stringify(ast, null, 4))
// Deno.writeTextFileSync('code.txt', compiler.instructions.map(i => `${i.opcode}${i.args.length > 0 ? ' ' : ''}${i.args.join(',')}`).join('\n'))